__author__ = 'Oleg Butovich'
__copyright__ = '(c) Oleg Butovich 2013-2017'
__licence__ = 'MIT'

import importlib
import posixpath
import logging

# Python 3 compatibility:
try:
    import httplib
except ImportError:  # py3
    from http import client as httplib
try:
    import urlparse
except ImportError:  # py3
    from urllib import parse as urlparse
try:
    basestring
except NameError:  # py3
    basestring = (bytes, str)

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARNING)


# https://metacpan.org/pod/AnyEvent::HTTP
ANYEVENT_HTTP_STATUS_CODES = {
    595: "Errors during connection establishment, proxy handshake",
    596: "Errors during TLS negotiation, request sending and header processing",
    597: "Errors during body receiving or processing",
    598: "User aborted request via on_header or on_body",
    599: "Other, usually nonretryable, errors (garbled URL etc.)"
}

SERVICES = {
    "PVE": {"supported_backends": ["https", "openssh", "ssh_paramiko"], "supported_https_auths": ["password", "token"], "default_port": 8006, "token_separator": "=", "ssh_additional_options": "--output-format json"},
    "PMG": {"supported_backends": ["https", "openssh", "ssh_paramiko"], "supported_https_auths": ["password"], "default_port": 8006},
    "PBS": {"supported_backends": ["https"], "supported_https_auths": ["password", "token"], "default_port": 8007, "token_separator": ":"}}

def config_failure(message, *args):
    raise NotImplementedError(message.format(*args))

class ProxmoxResourceBase(object):

    def __getattr__(self, item):
        if item.startswith("_"):
            raise AttributeError(item)

        kwargs = self._store.copy()
        kwargs['base_url'] = self.url_join(self._store["base_url"], item)

        return ProxmoxResource(**kwargs)

    def url_join(self, base, *args):
        scheme, netloc, path, query, fragment = urlparse.urlsplit(base)
        path = path if len(path) else "/"
        path = posixpath.join(path, *[('%s' % x) for x in args])
        return urlparse.urlunsplit([scheme, netloc, path, query, fragment])


class ResourceException(Exception):
    def __init__(self, status_code, status_message, content, errors=None):
        self.status_code = status_code
        self.status_message = status_message
        self.content = content
        self.errors = errors
        if errors != None:
            content += " - {0}".format(errors)
        message = "{0} {1}: {2}".format(status_code, status_message, content).strip()
        super(ResourceException, self).__init__(message)


class ProxmoxResource(ProxmoxResourceBase):

    def __init__(self, **kwargs):
        self._store = kwargs

    def __call__(self, resource_id=None):
        if not resource_id:
            return self

        if isinstance(resource_id, basestring):
            resource_id = resource_id.split("/")
        elif not isinstance(resource_id, (tuple, list)):
            resource_id = [str(resource_id)]

        kwargs = self._store.copy()
        if resource_id is not None:
            kwargs["base_url"] = self.url_join(self._store["base_url"], *resource_id)

        return ProxmoxResource(**kwargs)

    def _request(self, method, data=None, params=None):
        url = self._store["base_url"]
        if data:
            logger.info('%s %s %r', method, url, data)
        else:
            logger.info('%s %s', method, url)
        resp = self._store["session"].request(method, url, data=data or None, params=params)
        logger.debug('Status code: %s, output: %s', resp.status_code, resp.content)

        if resp.status_code >= 400:
            if hasattr(resp, 'reason'):
                raise ResourceException(
                    resp.status_code,
                    httplib.responses.get(resp.status_code,
                                        ANYEVENT_HTTP_STATUS_CODES.get(resp.status_code)),
                    resp.reason,
                    (self._store["serializer"].loads(resp) or {}).get('errors')
                )
            else:
                raise ResourceException(
                    resp.status_code,
                    httplib.responses.get(resp.status_code,
                                          ANYEVENT_HTTP_STATUS_CODES.get(resp.status_code)),
                    resp.text
                )
        elif 200 <= resp.status_code <= 299:
            return self._store["serializer"].loads(resp)

    def get(self, *args, **params):
        return self(args)._request("GET", params=params)

    def post(self, *args, **data):
        return self(args)._request("POST", data=data)

    def put(self, *args, **data):
        return self(args)._request("PUT", data=data)

    def delete(self, *args, **params):
        return self(args)._request("DELETE", params=params)

    def create(self, *args, **data):
        return self.post(*args, **data)

    def set(self, *args, **data):
        return self.put(*args, **data)


class ProxmoxAPI(ProxmoxResource):
    def __init__(self, host, backend='https', service='PVE', **kwargs):
        service = service.upper()

        # throw error for unsupported services
        if not service in SERVICES.keys():
            config_failure("{} service is not supported", service)

        # throw error for unsupported backend for service
        if not backend in SERVICES[service]["supported_backends"]:
            config_failure("{} does not support {} backend", service, backend)

        #load backend module
        self._backend = importlib.import_module('.backends.%s' % backend, 'proxmoxer').Backend(host, service=service, **kwargs)
        self._backend_name = backend

        self._store = {
            "base_url": self._backend.get_base_url(),
            "session": self._backend.get_session(),
            "serializer": self._backend.get_serializer(),
        }

    def get_tokens(self):
        """Return the auth and csrf tokens.

        Returns (None, None) if the backend is not https using password authentication.
        """
        if self._backend_name != 'https':
            return None, None

        return self._backend.get_tokens()



__author__ = 'Oleg Butovich'
__copyright__ = '(c) Oleg Butovich 2013-2017'
__licence__ = 'MIT'

from itertools import islice

try:
    import itertools.izip as zip
except ImportError:
    pass

from nose.tools import eq_, ok_, raises
from proxmoxer.core import ResourceException

class BaseSSHSuite(object):
    proxmox = None
    client = None
    session = None

    def __init__(self, sudo=False):
        self.sudo = sudo

    def _split_cmd(self, cmd):
        splitted = cmd.split()
        if not self.sudo:
            eq_(splitted[0], 'pvesh')
        else:
            eq_(splitted[0], 'sudo')
            eq_(splitted[1], 'pvesh')
            splitted.pop(0)
        options_set = set((' '.join((k, v)) for k, v in
                           zip(islice(splitted, 3, None, 2),
                               islice(splitted, 4, None, 2))))
        return ' '.join(splitted[1:3]), options_set

    def _get_called_cmd(self):
        raise NotImplementedError()

    def _set_stdout(self, stdout):
        raise NotImplementedError()

    def _set_stderr(self, stderr):
        raise NotImplementedError()

    def test_get(self):
        self._set_stdout("""
            [
               {
                  "subdir" : "status"
               },
               {
                  "subdir" : "content"
               },
               {
                  "subdir" : "upload"
               },
               {
                  "subdir" : "rrd"
               },
               {
                  "subdir" : "rrddata"
               }
            ]""")
        result = self.proxmox.nodes('proxmox').storage('local').get()
        eq_(self._get_called_cmd(), self._called_cmd('pvesh get /nodes/proxmox/storage/local --output-format json'))
        eq_(result[0]['subdir'], 'status')
        eq_(result[1]['subdir'], 'content')
        eq_(result[2]['subdir'], 'upload')
        eq_(result[3]['subdir'], 'rrd')
        eq_(result[4]['subdir'], 'rrddata')

    def test_delete(self):
        self.proxmox.nodes('proxmox').openvz(100).delete()
        eq_(self._get_called_cmd(), self._called_cmd('pvesh delete /nodes/proxmox/openvz/100 --output-format json'))
        self._set_stderr("200 OK")
        self.proxmox.nodes('proxmox').openvz('101').delete()
        eq_(self._get_called_cmd(), self._called_cmd('pvesh delete /nodes/proxmox/openvz/101 --output-format json'))
        self._set_stderr("200 OK")
        self.proxmox.nodes('proxmox').openvz.delete('102')
        eq_(self._get_called_cmd(), self._called_cmd('pvesh delete /nodes/proxmox/openvz/102 --output-format json'))

    def test_post(self):
        self._set_stderr("200 OK")
        node = self.proxmox.nodes('proxmox')
        node.openvz.create(vmid=800,
                           ostemplate='local:vztmpl/debian-6-turnkey-core_12.0-1_i386.tar.gz',
                           hostname='test',
                           storage='local',
                           memory=512,
                           swap=512,
                           cpus=1,
                           disk=4,
                           password='secret',
                           ip_address='10.0.100.222')
        cmd, options = self._split_cmd(self._get_called_cmd())
        eq_(cmd, 'create /nodes/proxmox/openvz')
        ok_("-cpus '1'" in options)
        ok_("-disk '4'" in options)
        ok_("-hostname 'test'" in options)
        ok_("-ip_address '10.0.100.222'" in options)
        ok_("-memory '512'" in options)
        ok_("-ostemplate 'local:vztmpl/debian-6-turnkey-core_12.0-1_i386.tar.gz'" in options)
        ok_("-password 'secret'" in options)
        ok_("-storage 'local'" in options)
        ok_("-swap '512'" in options)
        ok_("-vmid '800'" in options)

        self._set_stderr("200 OK")
        node = self.proxmox.nodes('proxmox1')
        node.openvz.post(vmid=900,
                         ostemplate='local:vztmpl/debian-7-turnkey-core_12.0-1_i386.tar.gz',
                         hostname='test1',
                         storage='local1',
                         memory=1024,
                         swap=1024,
                         cpus=2,
                         disk=8,
                         password='secret1',
                         ip_address='10.0.100.111')
        cmd, options = self._split_cmd(self._get_called_cmd())
        eq_(cmd, 'create /nodes/proxmox1/openvz')
        ok_("-cpus '2'" in options)
        ok_("-disk '8'" in options)
        ok_("-hostname 'test1'" in options)
        ok_("-ip_address '10.0.100.111'" in options)
        ok_("-memory '1024'" in options)
        ok_("-ostemplate 'local:vztmpl/debian-7-turnkey-core_12.0-1_i386.tar.gz'" in options)
        ok_("-password 'secret1'" in options)
        ok_("-storage 'local1'" in options)
        ok_("-swap '1024'" in options)
        ok_("-vmid '900'" in options)

    def test_put(self):
        self._set_stderr("200 OK")
        node = self.proxmox.nodes('proxmox')
        node.openvz(101).config.set(cpus=4, memory=1024, ip_address='10.0.100.100', onboot=True)
        cmd, options = self._split_cmd(self._get_called_cmd())
        eq_(cmd, 'set /nodes/proxmox/openvz/101/config')
        ok_("-memory '1024'" in options)
        ok_("-ip_address '10.0.100.100'" in options)
        ok_("-onboot 'True'" in options)
        ok_("-cpus '4'" in options)

        self._set_stderr("200 OK")
        node = self.proxmox.nodes('proxmox1')
        node.openvz('102').config.put(cpus=2, memory=512, ip_address='10.0.100.200', onboot=False)
        cmd, options = self._split_cmd(self._get_called_cmd())
        eq_(cmd, 'set /nodes/proxmox1/openvz/102/config')
        ok_("-memory '512'" in options)
        ok_("-ip_address '10.0.100.200'" in options)
        ok_("-onboot 'False'" in options)
        ok_("-cpus '2'" in options)

    @raises(ResourceException)
    def test_error(self):
        self._set_stderr("500 whoops")
        self.proxmox.nodes('proxmox').get()

    def test_no_error_with_extra_output(self):
        self._set_stderr("Extra output\n200 OK")
        self.proxmox.nodes('proxmox').get()

    @raises(ResourceException)
    def test_error_with_extra_output(self):
        self._set_stderr("Extra output\n500 whoops")
        self.proxmox.nodes('proxmox').get()

    def _called_cmd(self, cmd):
        called_cmd = cmd
        if self.sudo:
            called_cmd = 'sudo ' + cmd
        return called_cmd



__author__ = 'Oleg Butovich'
__copyright__ = '(c) Oleg Butovich 2013-2017'
__licence__ = 'MIT'


import json
import os
import sys
import time
import logging
from proxmoxer.core import SERVICES, config_failure

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARNING)

STREAMING_SIZE_THRESHOLD = 100 * 1024 * 1024 # 10 MiB

try:
    import requests
    urllib3 = requests.packages.urllib3
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    from requests.auth import AuthBase
    from requests.cookies import cookiejar_from_dict
except ImportError:
    logger.error("Chosen backend requires 'requests' module\n")
    sys.exit(1)

if sys.version_info[0] >= 3:
    import io
    def is_file(obj): return isinstance(obj, io.IOBase)
    # prefer using monoatomic time if available
    def get_time(): return time.monotonic()
else:
    def is_file(obj): return isinstance(obj, file)
    def get_time(): return time.time()


class AuthenticationError(Exception):
    def __init__(self, msg):
        super(AuthenticationError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg

    def __repr__(self):
        return self.__str__()


class ProxmoxHTTPAuthBase(AuthBase):

    def get_cookies(self):
        return cookiejar_from_dict({})

    def get_tokens(self):
        return None, None


class ProxmoxHTTPAuth(ProxmoxHTTPAuthBase):
    # number of seconds between renewing access tickets (must be less than 7200 to function correctly)
    # if calls are made less frequently than 2 hrs, using the API token auth is reccomended
    renew_age = 3600

    def __init__(self, base_url, username, password, otp=None, verify_ssl=False, timeout=5, service='PVE'):
        self.base_url = base_url
        self.username = username
        self.verify_ssl = verify_ssl
        self.timeout = timeout
        self.service = service
        self.pve_auth_ticket = ""

        self._getNewTokens(password=password, otp=otp)


    def _getNewTokens(self, password=None, otp=None):
        if password == None:
            # refresh from existing (unexpired) ticket
            password = self.pve_auth_ticket

        data = {"username": self.username, "password": password}
        if otp:
            data["otp"] = otp

        response_data = requests.post(self.base_url + "/access/ticket",
                                      verify=self.verify_ssl,
                                      timeout=self.timeout,
                                      data=data).json()["data"]
        if response_data is None:
            raise AuthenticationError("Couldn't authenticate user: {0} to {1}".format(self.username, self.base_url + "/access/ticket"))

        self.birth_time = get_time()
        self.pve_auth_ticket = response_data["ticket"]
        self.csrf_prevention_token = response_data["CSRFPreventionToken"]

    def get_cookies(self):
        return cookiejar_from_dict({self.service + "AuthCookie": self.pve_auth_ticket})

    def get_tokens(self):
        return self.pve_auth_ticket, self.csrf_prevention_token

    def __call__(self, r):
        #refresh ticket if older than `renew_age`
        if (get_time() - self.birth_time) >= self.renew_age:
            logger.debug("refreshing ticket (age {0})".format(get_time() - self.birth_time))
            self._getNewTokens()

        # only attach CSRF token if needed (reduce interception risk)
        if r.method != 'GET':
            r.headers["CSRFPreventionToken"] = self.csrf_prevention_token
        return r


# DEPRECATED(1.1.0) - either use a password or the API Tokens
class ProxmoxHTTPTicketAuth(ProxmoxHTTPAuth):
    """Use existing ticket/token to create a session.

    Overrides ProxmoxHTTPAuth so that an existing auth ticket and csrf token
    may be used instead of passing username/password.
    """
    def __init__(self, auth_ticket, csrf_token):
        self.pve_auth_ticket = auth_ticket
        self.csrf_prevention_token = csrf_token
        self.birth_time = get_time()

        # deprecation notice
        logger.warning("** Existing token auth is Deprecated as of 1.1.0\n** Please use the API Token Auth for long-running programs or pass existing ticket as password to the user/password auth")


class ProxmoxHTTPApiTokenAuth(ProxmoxHTTPAuthBase):
    def __init__(self, username, token_name, token_value, service):
        self.service = service
        self.username = username
        self.token_name = token_name
        self.token_value = token_value

    def __call__(self, r):
        r.headers["Authorization"] = "{0}APIToken={1}!{2}{3}{4}".format(self.service, self.username, self.token_name, SERVICES[self.service]["token_separator"], self.token_value)
        return r


class JsonSerializer(object):
    content_types = [
        "application/json",
        "application/x-javascript",
        "text/javascript",
        "text/x-javascript",
        "text/x-json"
        ]

    def get_accept_types(self):
        return ", ".join(self.content_types)

    def loads(self, response):
        try:
            return json.loads(response.content.decode('utf-8'))['data']
        except (UnicodeDecodeError, ValueError):
            return {"errors": response.content}


class ProxmoxHttpSession(requests.Session):

    def request(self, method, url, params=None, data=None, headers=None, cookies=None, files=None, auth=None,
                timeout=None, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None,
                serializer=None):

        a = auth or self.auth
        c = cookies or self.cookies

        # take set verify flag from session if request does not have this parameter explicitly
        if verify is None:
            verify = self.verify

        # pull cookies from auth if not present
        if (not c) and a:
            cookies = a.get_cookies()

        # filter out streams
        files = files or {}
        data = data or {}
        isLargePayload = False
        totalFileSize = 0
        for k, v in data.copy().items():
            if is_file(v):
                totalFileSize += getFileSize(v)
                if totalFileSize > STREAMING_SIZE_THRESHOLD:
                    isLargePayload = True
                
                # add in filename from file pointer (patch for https://github.com/requests/toolbelt/pull/316)
                files[k] = (requests.utils.guess_filename(v), v)
                del data[k]

        # if there are any large file, send all data and files using streaming multipart encoding
        if isLargePayload:
            try:
                from requests_toolbelt import MultipartEncoder
                encoder = MultipartEncoder(fields=mergeDicts(data, files))
                data = encoder
                files = None
                headers = {'Content-Type': encoder.content_type}
            except ImportError:
                # if the files will cause issues with the SSL 2GiB limit (https://bugs.python.org/issue42853#msg384566)
                if totalFileSize > 2147483135: #2^31 - 1 - 512
                    logger.warn(
                        "Install 'requests_toolbelt' to add support for files larger than 2GiB")
                    raise OverflowError("Unable to upload a payload larger than 2 GiB")
                else:
                    logger.info("Installing 'requests_toolbelt' will deacrease memory used during upload")


        if not files and serializer:
            headers = {"Content-Type": 'application/x-www-form-urlencoded'}

        return super(ProxmoxHttpSession, self).request(method, url, params, data, headers, cookies, files, auth,
                                                       timeout, allow_redirects, proxies, hooks, stream, verify, cert)


class Backend(object):
    def __init__(self, host, user=None, password=None, otp=None, port=None,
                 verify_ssl=True, mode='json', timeout=5, auth_token=None,
                 csrf_token=None, token_name=None, token_value=None, service='PVE'):
        if ':' in host:
            host, host_port = host.split(':')
            port = host_port if host_port.isdigit() else port

        # if a port is not specified, use the default port for this service
        if not port:
            port = SERVICES[service]["default_port"]

        self.base_url = "https://{0}:{1}/api2/{2}".format(host, port, mode)

        if auth_token is not None:
            # DEPRECATED(1.1.0) - either use a password or the API Tokens
            self.auth = ProxmoxHTTPTicketAuth(auth_token, csrf_token)
        elif token_name is not None:
            if not "token" in SERVICES[service]["supported_https_auths"]:
                config_failure("{} does not support API Token authentication", service)

            self.auth = ProxmoxHTTPApiTokenAuth(user, token_name, token_value, service)
        elif password is not None:
            if not "password" in SERVICES[service]["supported_https_auths"]:
                config_failure("{} does not support password authentication", service)

            self.auth = ProxmoxHTTPAuth(self.base_url, user, password, otp, verify_ssl, timeout, service)
        self.verify_ssl = verify_ssl
        self.mode = mode
        self.timeout = timeout

    def get_session(self):
        session = ProxmoxHttpSession()
        session.verify = self.verify_ssl
        session.auth = self.auth
        # cookies are taken from the auth
        session.headers['Connection'] = 'keep-alive'
        session.headers["accept"] = self.get_serializer().get_accept_types()
        return session

    def get_base_url(self):
        return self.base_url

    def get_serializer(self):
        assert self.mode == 'json'
        return JsonSerializer()

    def get_tokens(self):
        """Return the in-use auth and csrf tokens if using user/password auth."""
        return self.auth.get_tokens()


def getFileSize(fileObj):
    # store existing file cursor location
    startingCursor = fileObj.tell()

    # get size
    size = fileObj.seek(0, os.SEEK_END)

    # reset cursor
    fileObj.seek(startingCursor)

    return size


def mergeDicts(*dicts):
    # compatibility function for missing unpack operator
    # synonymous with {**dict for dict in dicts}
    return {k: v for d in dicts for k, v in d.items()}



__author__ = 'Oleg Butovich'
__copyright__ = '(c) Oleg Butovich 2013-2017'
__licence__ = 'MIT'

import io
from mock import patch
from nose.tools import eq_, assert_raises
from proxmoxer import ProxmoxAPI
from tests.base.base_ssh_suite import BaseSSHSuite


@patch('paramiko.SSHClient')
def test_paramiko_connection(_):
    proxmox = ProxmoxAPI('proxmox', user='root', backend='ssh_paramiko', port=123)
    session = proxmox._store['session']
    eq_(session.ssh_client.connect.call_args[0], ('proxmox',))
    eq_(session.ssh_client.connect.call_args[1], {'username': 'root',
                                                  'allow_agent': True,
                                                  'key_filename': None,
                                                  'look_for_keys': True,
                                                  'timeout': 5,
                                                  'password': None,
                                                  'port': 123})


@patch('paramiko.SSHClient')
def test_paramiko_invalid_backend(_):
    with assert_raises(NotImplementedError):
        ProxmoxAPI('proxmox', user='root', backend='ssh_paramiko', port=123, service='PBS')



@patch('paramiko.SSHClient')
def test_paramiko_tokens(_):
        p = ProxmoxAPI('proxmox', user='root', backend='ssh_paramiko', port=123)
        eq_(p.get_tokens()[0], None)
        eq_(p.get_tokens()[1], None)


class TestParamikoSuite(BaseSSHSuite):

    # noinspection PyMethodOverriding
    @patch('paramiko.SSHClient')
    def setUp(self, _):
        self.proxmox = ProxmoxAPI('proxmox', user='root', backend='ssh_paramiko', port=123)
        self.client = self.proxmox._store['session'].ssh_client
        self.session = self.client.get_transport().open_session()
        self._set_stderr('200 OK')
        self._set_stdout('')

    def _get_called_cmd(self):
        return self.session.exec_command.call_args[0][0]

    def _set_stdout(self, stdout):
        self.session.makefile.return_value = io.BytesIO(stdout.encode('utf-8'))

    def _set_stderr(self, stderr):
        self.session.makefile_stderr.return_value = io.BytesIO(stderr.encode('utf-8'))


class TestParamikoSuiteWithSudo(BaseSSHSuite):

    # noinspection PyMethodOverriding
    @patch('paramiko.SSHClient')
    def setUp(self, _):
        super(TestParamikoSuiteWithSudo, self).__init__(sudo=True)
        self.proxmox = ProxmoxAPI('proxmox', user='root', backend='ssh_paramiko', port=123, sudo=True)
        self.client = self.proxmox._store['session'].ssh_client
        self.session = self.client.get_transport().open_session()
        self._set_stderr('200 OK')
        self._set_stdout('')

    def _get_called_cmd(self):
        return self.session.exec_command.call_args[0][0]

    def _set_stdout(self, stdout):
        self.session.makefile.return_value = io.BytesIO(stdout.encode('utf-8'))

    def _set_stderr(self, stderr):
        self.session.makefile_stderr.return_value = io.BytesIO(stderr.encode('utf-8'))



__author__ = 'Oleg Butovich'
__copyright__ = '(c) Oleg Butovich 2013-2017'
__licence__ = 'MIT'


from itertools import chain
import json
import re
import logging
import sys
from proxmoxer.core import SERVICES

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARNING)

class Response(object):
    def __init__(self, content, status_code):
        self.status_code = status_code
        self.content = content
        self.text = str(content)
        self.headers = {"content-type": "application/json"}


class ProxmoxBaseSSHSession(object):

    def _exec(self, cmd):
        raise NotImplementedError()

    # noinspection PyUnusedLocal
    def request(self, method, url, data=None, params=None, headers=None):
        method = method.lower()
        data = data or {}
        params = params or {}
        url = url.strip()

        cmd = {'post': 'create',
               'put': 'set'}.get(method, method)

        # for 'upload' call some workaround
        tmp_filename = ''
        if url.endswith('upload'):
            # copy file to temporary location on proxmox host
            tmp_filename, _ = self._exec(
                "python -c 'import tempfile; import sys; tf = tempfile.NamedTemporaryFile(); sys.stdout.write(tf.name)'")
            self.upload_file_obj(data['filename'], tmp_filename)
            data['filename'] = data['filename'].name
            data['tmpfilename'] = tmp_filename

        translated_data = ' '.join(["-{0} '{1}'".format(k, v) for k, v in chain(data.items(), params.items())])

        additional_options = SERVICES[self.service.upper()].get("ssh_additional_options", "")
        full_cmd = '{0}sh {1} {2}'.format(self.service, ' '.join(filter(None, (cmd, url, translated_data))), additional_options)

        stdout, stderr = self._exec(full_cmd)
        def match(s): return re.match(r'\d\d\d [a-zA-Z]', s)
        if stderr:
            # sometimes contains extra text like 'trying to acquire lock...OK'
            status_code = next(
                (int(s.split()[0]) for s in stderr.splitlines() if match(s)),
                500)
        else:
            status_code = 200
        if stdout:
            return Response(stdout, status_code)
        return Response(stderr, status_code)

    def upload_file_obj(self, file_obj, remote_path):
        raise NotImplementedError()


class JsonSimpleSerializer(object):

    def loads(self, response):
        try:
            return json.loads(response.content)
        except (UnicodeDecodeError, ValueError):
            return {"errors": response.content}


class BaseBackend(object):

    def get_session(self):
        return self.session

    def get_base_url(self):
        return ''

    def get_serializer(self):
        return JsonSimpleSerializer()
